from collections import namedtuple
from typing import Any

from django.db import models
from rest_framework import serializers
from rest_framework.fields import empty

from django_features.custom_fields.models import CustomField
from django_features.custom_fields.models import CustomFieldBaseModel
from django_features.custom_fields.models import CustomValue


class CustomChoiceSerializer(serializers.ModelSerializer):
    class Meta:
        model = CustomValue
        fields = ["id", "label", "value"]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        if isinstance(self.instance, CustomValue):
            field = self.instance.field
            self.fields["value"] = CustomField.TYPE_SERIALIZER_MAP[field.field_type](
                allow_null=True, read_only=True, required=False
            )


class CustomFieldSerializer(serializers.ModelSerializer):
    choices = serializers.SerializerMethodField()

    class Meta:
        model = CustomField
        fields = [
            "choice_field",
            "choices",
            "created",
            "editable",
            "external_key",
            "field_type",
            "hidden",
            "id",
            "identifier",
            "label",
            "modified",
            "multiple",
            "order",
        ]

    def get_choices(self, obj: CustomField) -> list:
        return CustomChoiceSerializer(obj.choices, many=True).data


CustomFieldData = namedtuple(
    "CustomFieldData",
    [
        "id",
        "identifier",
        "choices",
        "choice_field",
        "multiple",
        "serializer_field",
    ],
)


class CustomFieldBaseModelSerializer(serializers.ModelSerializer):
    _custom_fields: list[CustomFieldData] = []
    _unique_choice_field = "id"
    _write_only_serializer = False

    class Meta:
        abstract = True
        fields = "__all__"
        model = None

    def __init__(
        self,
        instance: Any = None,
        data: Any = empty,
        exclude_custom_fields: bool = False,
        write_only_serializer: bool = False,
        **kwargs: Any,
    ) -> None:
        self.exclude_custom_fields: bool = exclude_custom_fields
        self.write_only_serializer = (
            write_only_serializer or self._write_only_serializer
        )
        super().__init__(instance, data, **kwargs)

    @property
    def model(self) -> models.Model:
        if not self.Meta.model:
            raise ValueError("Meta.model must be set")
        return self.Meta.model

    def get_fields(self) -> dict[str, Any]:
        fields = super().get_fields()
        if self.exclude_custom_fields:
            return fields
        self._custom_fields = []
        custom_fields = list(CustomField.objects.for_model(self.model))
        for field in custom_fields:
            self._custom_fields.append(
                CustomFieldData(
                    field.id,
                    field.identifier,
                    field.choices,
                    field.choice_field,
                    field.multiple,
                    field.serializer_field,
                )
            )
            serialized_field = field.serializer_field
            if field.choice_field:
                serialized_field.set_unique_field(self._unique_choice_field)
            fields[field.identifier] = serialized_field
        return fields

    def create(self, validated_data: dict) -> Any:
        custom_value_instances: list[CustomValue] = []
        choices: list[CustomValue] = []
        for field in self._custom_fields:
            value = validated_data.pop(field.identifier, None)
            if value is None:
                continue
            if not field.choice_field:
                custom_value_instances.append(
                    CustomValue(
                        field_id=field.id,
                        value=self.fields[field.identifier].to_representation(value),
                    )
                )
            else:
                if field.multiple:
                    choices.extend(value)
                else:
                    choices.append(value)
        instance = super().create(validated_data)
        if custom_value_instances or choices:
            custom_values = CustomValue.objects.bulk_create(custom_value_instances)
            custom_values.extend(choices)
            instance.custom_values.set(custom_values)
        return instance

    def _create_or_update_custom_value(
        self, instance: CustomFieldBaseModel, field: CustomFieldData, value: Any
    ) -> None:
        serializer_field = field.serializer_field
        serializer_field.run_validators(value)
        if value is not None:
            value = serializer_field.to_representation(value)
        try:
            value_object = instance.custom_values.select_related("field").get(
                field_id=field.id
            )
            if value is None:
                value_object.delete()
            else:
                value_object.value = value
                value_object.save()
        except CustomValue.DoesNotExist:
            if value is not None:
                value_object = CustomValue.objects.create(
                    field_id=field.id, value=value
                )
                instance.custom_values.add(value_object)

    def update(self, instance: Any, validated_data: dict) -> Any:
        for field in self._custom_fields:
            if field.identifier not in validated_data:
                continue
            value = validated_data.pop(field.identifier, None)
            if field.choice_field:
                instance.custom_values.remove(*field.choices)
                if value is None:
                    continue
                if field.multiple:
                    instance.custom_values.add(*value)
                else:
                    instance.custom_values.add(value)
            else:
                self._create_or_update_custom_value(instance, field, value)
        instance = super().update(instance, validated_data)
        return instance



__all__ = [
    "CustomFieldBaseModel",
    "CustomFieldTypeBaseModel",
    "CustomField",
    "CustomValue",
]


from .base import CustomFieldBaseModel
from .base import CustomFieldTypeBaseModel
from .field import CustomField
from .value import CustomValue



from typing import Any

from django.core.exceptions import ObjectDoesNotExist
from pluck import pluck
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from rest_framework.utils.model_meta import get_field_info

from django_features.custom_fields.models import CustomField
from django_features.custom_fields.models import CustomValue
from django_features.custom_fields.models.value import CustomValueQuerySet
from django_features.custom_fields.serializers import CustomChoiceSerializer


class ChoiceIdField(serializers.Field):
    _unique_field: str | None = None

    def __init__(
        self, field: CustomField, unique_field: str | None = None, **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        self.field = field
        self.required = kwargs.get("required", self.field.required)
        self.set_unique_field(unique_field)

    def set_unique_field(self, unique_field: str | None) -> None:
        self._unique_field = unique_field or "id"

        valid_fields = get_field_info(CustomValue).fields_and_pk
        if self._unique_field not in valid_fields:
            raise ValueError(
                f"The unique_field must be a valid field of {valid_fields}: invalid field {self.unique_field}"
            )

    def get_queryset(self) -> CustomValueQuerySet:
        return CustomValue.objects.filter(field_id=self.field.id)

    def to_representation(
        self, value: CustomValue | CustomValueQuerySet
    ) -> int | list[int]:
        return CustomChoiceSerializer(value, many=self.field.multiple).data

    def _choice_field(self, data: int | str | dict) -> CustomValue:
        if isinstance(data, dict):
            value = data.get("id")
        else:
            value = data
        try:
            return self.get_queryset().get(**{self._unique_field: value})
        except ObjectDoesNotExist:
            raise ValidationError(
                f"Custom value with the {self._unique_field} {data} does not exist."
            )

    def _multiple_choice(self, data: list[int | str | dict]) -> CustomValueQuerySet:
        if all(type(d) is dict for d in data):
            value = pluck(data, self._unique_field)
        else:
            value = data
        values = self.get_queryset().filter(**{f"{self._unique_field}__in": value})
        if values.count() != len(data):
            missing = set(data) - set(values.values_list(self._unique_field, flat=True))
            raise ValidationError(
                f"Some of the given {self._unique_field}s do not match: {missing}"
            )
        return values

    def to_internal_value(self, data: Any) -> CustomValue | CustomValueQuerySet:
        if self.field.multiple and isinstance(data, list):
            return self._multiple_choice(data)
        elif self.field.choice_field and isinstance(data, (int, str, dict)):
            return self._choice_field(data)
        else:
            if not self.field.choice_field:
                raise ValidationError(f"The field {self.field} is not a choice field.")
            raise ValidationError(
                f"The given value {data} has not a valid type. Expected a list or int."
            )



from rest_framework.routers import DefaultRouter

from django_features.system_message.viewsets import SystemMessageTypeViewSet
from django_features.system_message.viewsets import SystemMessageViewSet


system_message_router = DefaultRouter(trailing_slash=False)
system_message_router.register(r"system_message", SystemMessageViewSet)
system_message_router.register(r"system_message_type", SystemMessageTypeViewSet)



from typing import Any

from django.db.models import Q
from django.db.models import QuerySet
from django.utils import timezone
from django_filters import rest_framework as filters
from django_filters.rest_framework import DjangoFilterBackend
from django_filters.widgets import QueryArrayWidget
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet
from rest_framework.viewsets import ReadOnlyModelViewSet

from django_features.pagination import PageNumberPaginator
from django_features.system_message import models
from django_features.system_message import serializers
from django_features.system_message.permissions import CanManageSystemMessage


class SystemMessageTypeViewSet(ReadOnlyModelViewSet):
    queryset = models.SystemMessageType.objects.all()
    serializer_class = serializers.SystemMessageTypeSerializer


class SystemMessageFilter(filters.FilterSet):
    active = filters.BooleanFilter(method="filter_active")
    dismissed = filters.BooleanFilter(method="filter_dismissed")
    type = filters.BaseInFilter(field_name="type", widget=QueryArrayWidget)

    class Meta:
        model = models.SystemMessage
        fields = ["id", "type"]

    def filter_active(self, queryset: QuerySet, name: str, value: bool) -> QuerySet:
        q = Q(begin__lte=timezone.now()) & (
            Q(end__gte=timezone.now()) | Q(end__isnull=True)
        )
        if not value:
            return queryset.exclude(q)
        return queryset.filter(q)

    def filter_dismissed(self, queryset: QuerySet, name: str, value: bool) -> QuerySet:
        if not self.request.user.is_authenticated:
            return queryset
        q = Q(dismissed_users=self.request.user)
        if not value:
            return queryset.exclude(q)
        return queryset.filter(q)


class SystemMessageViewSet(ModelViewSet):
    filter_backends = (OrderingFilter, DjangoFilterBackend, SearchFilter)
    filterset_class = SystemMessageFilter
    ordering_fields = (
        "background_color",
        "begin",
        "end",
        "order",
        "text",
        "text_color",
        "title",
        "type",
    )
    pagination_class = PageNumberPaginator
    permission_classes = [CanManageSystemMessage]
    queryset = models.SystemMessage.objects.all()
    search_fields = ("background_color", "message", "message_color", "title", "type")
    serializer_class = serializers.SystemMessageSerializer

    @action(methods=["patch"], detail=True)
    def dismiss(self, request: Request, pk: int, *args: Any, **kwargs: Any) -> Response:
        if self.request.user.is_authenticated:
            self.get_object().dismissed_users.add(self.request.user)
        return Response(status=status.HTTP_204_NO_CONTENT)



# Generated by Django 4.2.23 on 2025-10-14 11:53

import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("custom_fields", "0002_make_custom_field_label_and_value_text_translateable"),
    ]

    operations = [
        migrations.AddField(
            model_name="customfield",
            name="type_content_type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="customfield_set_for_type",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AddField(
            model_name="customfield",
            name="type_id",
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
    ]



from typing import Any

from constance import config
from rest_framework.permissions import DjangoModelPermissions
from rest_framework.request import Request


class CanManageSystemMessage(DjangoModelPermissions):
    allowed_actions = ["list", "retrieve"]

    def has_permission(self, request: Request, view: Any) -> bool:
        if not config.ENABLE_SYSTEM_MESSAGE:
            return False
        if view.action in self.allowed_actions:
            return True
        perm = config.SYSTEM_MESSAGE_PERMISSION
        if not perm:
            return True
        return request.user.is_superuser or request.user.has_perm(perm)
